# Copyright 2007-present Greg Hurrell. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

require 'spec_helper'
require 'uri'

describe Wikitext, 'encoding a link target' do
  it 'should complain if passed nil' do
    lambda { Wikitext::Parser.encode_link_target(nil) }.should raise_error
  end

  it 'should do nothing on zero-length input' do
    Wikitext::Parser.encode_link_target('').should == ''
  end

  it 'should convert embedded spaces into "%20"' do
    Wikitext::Parser.encode_link_target('hello world').should == 'hello%20world'
  end

  it 'should eat leading spaces' do
    Wikitext::Parser.encode_link_target(' hello world').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('  hello world').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('   hello world').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('    hello world').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('     hello world').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('      hello world').should == 'hello%20world'
  end

  it 'should eat trailing spaces' do
    Wikitext::Parser.encode_link_target('hello world ').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('hello world  ').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('hello world   ').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('hello world    ').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('hello world     ').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('hello world      ').should == 'hello%20world'
  end

  it 'should eat leading and trailing spaces combined' do
    Wikitext::Parser.encode_link_target('     hello world ').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('     hello world ').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('    hello world  ').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('   hello world   ').should == 'hello%20world'
    Wikitext::Parser.encode_link_target('  hello world    ').should == 'hello%20world'
    Wikitext::Parser.encode_link_target(' hello world     ').should == 'hello%20world'
  end

  it 'should return nothing for input consisting entirely of spaces' do
    Wikitext::Parser.encode_link_target(' ').should == ''
    Wikitext::Parser.encode_link_target('  ').should == ''
    Wikitext::Parser.encode_link_target('   ').should == ''
    Wikitext::Parser.encode_link_target('    ').should == ''
    Wikitext::Parser.encode_link_target('     ').should == ''
    Wikitext::Parser.encode_link_target('      ').should == ''
  end

  it 'should convert reserved symbols into percent escapes' do
    Wikitext::Parser.encode_link_target('http://www.apple.com/q?foo').should == 'http%3a%2f%2fwww.apple.com%2fq%3ffoo'
  end

  it 'should convert non-ASCII into UTF-8 and then apply percent escapes' do
    Wikitext::Parser.encode_link_target('cañon').should == 'ca%c3%b1on'
  end

  it 'should handle mixed scenarios (commas, double-quotes and UTF-8)' do
    Wikitext::Parser.encode_link_target('foo, "bar" & baz €').should == 'foo%2c%20%22bar%22%20%26%20baz%20%e2%82%ac'
  end

  it 'should get the same answer as URI.escape' do
    Wikitext::Parser.encode_link_target('foo bar').should == 'foo%20bar'
    Wikitext::Parser.encode_link_target('http://www.google.com/search?q=hello&foo=bar').should == 'http%3a%2f%2fwww.google.com%2fsearch%3fq%3dhello%26foo%3dbar'
    Wikitext::Parser.encode_link_target('€').should == '%e2%82%ac'
  end
end
